------------------------------
------------------------------
--add_and_remove_functions.lua     << this file name
------------------------------
------------------------------

--This file contains functions to add and remove devices to/from the tags, both individually (tags) and full pages
------------------------------------------------------------------------------------------------------------------

--ADDING DEVICES
------------------
--1) add_selected_device_to_selected_tag()
--2) add_all_devices_in_track_to_tag(track_idx,tag_name) --defaults to selected track and tag but can be specified
--3) add_tracks_one_to_twelve_to_selected_page()

--REMOVING DEVICES
-------------------
--4) remove_entry_from_tab(page_idx,tab_idx,item_idx)



--SORTING DEVICES IN POPUP
--------------------------
--5) (helper fn) strip_out_device_name_from_display_name(display_name)
--5) sort_tag_list(target_page,target_tag) 


TYPE_INSTRUMENT = 1
TYPE_EFFECT = 2

--helpers:

--local save global object to xml (save file)
----------------------------
local function save_to_xml_file()
----------------------------
  --get xml bundle path and concatenate the file name
  local file_name = renoise.tool().bundle_path .. "saved_tags.xml"
  --save global document object to xml
  g_retrieved_xml:save_as(file_name)  
end

--5) Helper
--as we can't access short name for VST we need to get it out of the 'Display Name'
--this saves looping device_properties in the API.
--example output >>> 'VST: Kilohearts: kHs Transient Shaper' >> 'kHs Transient Shaper'
--------------------------------------------------------------------
function strip_out_device_name_from_display_name(display_name)
--------------------------------------------------------------------
  
  --NOTE: only pass from device in renoise and not plugin_info  which gives short name
  
  --counter to break while loop in case of problem
  local count = 1
  --loop up to 10 times
  while count < 10 do
    --find index of first colon
    local colon = string.find(display_name, ':+')
    --get substring from colon index to end of string
    -- (the '+ 2' is due to colons in display name having a space after them)
    if colon ~= nil then
      display_name = string.sub(display_name,colon + 2)
    else
      --no more colons found so return string
      return display_name
    end
   --increment counter (loop will break when counter reaches 10)
   count = count + 1
  end
end
------
-------------------------------------------------------


-------------------------------------------------------
--1)
--Add a single device to a tag
--function adds the selected device to the selected tag in the GUI -- renoise right click menu
----------------------------------------------
function add_selected_device_to_selected_tag()
----------------------------------------------
  
  if vb == nil then
    status("Gui Needs To Be Open") --TODO remove menu entry when gui closes
    return
  end
  
  --if a user page is not selected: status update and return
  if (g_selected_page ~= 2 and g_selected_page ~= 3) then
    error("A User Page Must Be Selected First")
    return
  end
  
  --song
  local song = renoise.song()
  --selected track device
  local new_dev = song.selected_track_device
  
  --error if no device/mixer device selected
  if new_dev == nil or song.selected_track_device_index == 1 then
    error("A Device Must Be Selected First In The Track")
    return
  end
  
  --new (selected) device name
  local new_name = new_dev.display_name
  --new plugin path
  local new_path = new_dev.device_path

  --get shortened display name
  new_name = strip_out_device_name_from_display_name(new_name)
  
  --declare fx pages[] node object
  local xml_page = g_retrieved_xml.plugin_types[TYPE_EFFECT].pages
  
  --loop xml paths to check if device has already been added, return if it has 
  for i = 1,#xml_page[g_selected_page].tags[g_selected_button].plugs do
    if new_dev.device_path == tostring(xml_page[g_selected_page].tags[g_selected_button].plugs[i].plug_path) then
      --status and return if already in tag
      status(new_name.." Was Already Added To This Tag")
      return
    end
  end
  
  --add device to xml node
  --insert_new_plugin_node(type,page,tag,name,path)
  insert_new_plugin_node(TYPE_EFFECT,g_selected_page,g_selected_button,new_name,new_path)
  --save the changed global object to xml
  save_to_xml_file()
  --update the main GUI
  --update_load_popup()  --should be done from calling function
  --status
  status("Device Added")
end
----eof


--2)
--Add all devices (OVERWRITE) in a track to a tag.
--default is selected track/tag, but can be specified with parameters (track_idx,tag_name)
--tag_name is a string "tag1","tag2",.."tag12"
---------------------------------------------------------------------
function add_all_devices_in_track_to_tag(track_idx,tag_idx,save_xml)
---------------------------------------------------------------------
  

 --TODO Check for entries being doubled etc
 --TODO update GUI after load operation
  --TODO OVERWRITE VS ADD?? 
 
  if vb == nil then
    status("Gui Needs To Be Open") --TODO remove menu entry when gui closes
    return
  end
  
  --if a user page is not selected: status update and return
  if (g_selected_page ~= 2 and g_selected_page ~= 3) then
    error("A User Page Must Be Selected First")
    return
  end

  --song
  local song = renoise.song()
  
  --use global for selected page
  local page = g_selected_page
  
  --track idx passed or default to selected track
  local track = nil
  if track_idx == nil then
    --if no track index is passed to the function, default to the selected track
    track = song.selected_track
  else --use specified track
    track = song:track(track_idx) 
  end
  
 
  --if tag index passed or default to global
  local tag = tag_idx or g_selected_button

  --create a table to add the device properties into.
  local device_props = {}
  device_props.name = {}
  device_props.path = {}
  
  --get fx pages[] node
  local xml_page = g_retrieved_xml.plugin_types[TYPE_EFFECT].pages 
  
   
  ---------------------------------------------
  --delete all current devices saved in the tag
  ---------------------------------------------
  for i = #xml_page[g_selected_page].tags[tag].plugs,1,-1 do
    --get name from xml object
    local name = tostring(xml_page[g_selected_page].tags[tag].plugs[i].plug_name)
    --delete using global function
    remove_plugin_node(TYPE_EFFECT,g_selected_page,tag,name)
  end
  
 
  --loop all plugs in track
  for idx = 2,#track.devices do
    --get existing/loaded device's name and path
    local dev = track:device(idx)
    local name = dev.display_name
    local path = dev.device_path
    local skip_plug = false
    --NOTE Not needed now as we delete xml entries above.  Would be needed if behaviour was to merge new entries
    ------------------------------------------------------------------------------------------------------------
    --loop xml plugin_path's and skip any plugs alerady added
    --for i = 1,#xml_page[g_selected_page].tags[tag].plugs do
    --  --is current device path found in xml already?
    --  if path == tostring(xml_page[page].tags[tag].plugs[i].plug_path.value) then --or 
    --    skip_plug = true
    --    break
    --  end
    --end
    --or has current path already been added to our table?
    for i = 1,#device_props.path do
      if path == device_props.path[i] then
        skip_plug = true
        break
      end
    end   
    --now only insert plug to table if it isn't already in xml or table
    if skip_plug == false then
      --shorten name to the short display name
      name = strip_out_device_name_from_display_name(name)
      --add name and path to table
      table.insert(device_props.name,name)
      table.insert(device_props.path,path)
    end
  end
  
  --return if no new plugs to be added to tag
  if device_props.name[1] == nil then
    status("All Plugs Were Already Present")
    return
  end
   

  --local function to sort the device_props table (called directly below definition)
  -----------------------------------------------
  local function sort_plugs_by_display_names(tab)
  -----------------------------------------------
    --temporary table
    local temp = {}
    --an append_digit will be added to the end of each entry after an index number.  The append_digit tells us how many
    --digits to expect in the index (normal usage should be 1/2 as unlikely that 100+ plugs would be added to a tag)
    --this is done so that if the end of a plugin name is a number, it will not confused with the added index in later sorting.
    --Example for table entry 7 after loop:
    --"[7] =>  "TR5 TAPE ECHO71" --- '7' is the index, '1' is the append_digit telling us that 7 is a single digit number.
    local append_digit = 1
    --loop entries, adding original table position index
    for i = 1,#tab.name do
      --set the append digit to the number of digits of 'i'
      append_digit = #tostring(i)
      --capitalise the plug name for sorting (keeps later sort results 'alphabetical' rather than ASCII)
      temp[i] = tab.name[i]:upper()
      --add the loop index number 'i' and + 'append_digit' to specify number of digits 'i' has (1 digit for 5, 2 digits for 16 etc.)
      temp[i] = temp[i]..tostring(i)..tostring(append_digit)
    end
    
    --sort the capitalised table alphabetically (Lua standard fn)
    table.sort(temp)
      
    --now the temporary table of plug names is sorted alphabetically we retrieve the
    --index numbers that were appended to the ends of the string entries denoting their original table positions
    ------------------------------------------------------------------------------------------------------------
    local index_table = {}
    for i = 1,#temp do
      --get append_digit
      local entry = temp[i]
      --get the append_digit which is the last character of the string.
      local append_digit = entry:sub(#entry)
      --get the index (appended number minus the append digit).  Add it to the index_table
      index_table[i] = entry:sub(#entry - append_digit, #entry-1)
    end
    
    --make a new 'sorted table' to hold the alphabetically sorted entries
    local sorted_tab = {}
    sorted_tab.name = {}
    sorted_tab.path = {}
    --loop to re-index original table
    for i = 1,#temp do
      --by indexing index_table we get the plug names back in alphabetical order.
      local index = tonumber(index_table[i])
      --now build the sorted table using the original saved indexes
      table.insert(sorted_tab.name,tab.name[index])
      table.insert(sorted_tab.path,tab.path[index])--make sure to add paths aswell
    end
    --return the sorted table
    return sorted_tab
  end 
  -------------------------- end local fn

  --sort the plugins alphabetically with above local function
  device_props = sort_plugs_by_display_names(device_props) 
  
  
   
  --update xml object
  for idx = 1,(#device_props.name)do 
    --get short name
    local name = strip_out_device_name_from_display_name(device_props.name[idx])
    --insert_new_plugin_node(type,page,tag,name,path)
    insert_new_plugin_node(TYPE_EFFECT,g_selected_page,tag,name,device_props.path[idx])
  end
 
  --save the changed global object to xml
  if save_xml == true then
    save_to_xml_file() 
  end
  --update the main GUI
  update_load_popup()
  --status
  status("Devices Added")
end--end fn
------------
------------
--add_all_devices_in_track_to_tag(track_idx,tag_name)

--3)
--Loads all the devices, track by track (1 to 12) from the renoise song
--This allows the user to save an xrns as a tag backup e.g. with all compressors on
--track 1, delays on track 2 ..etc.  The track name becomes the tag name (button names in gui) 
----------------------------------------------------
function add_tracks_one_to_twelve_to_selected_page()
----------------------------------------------------

  if vb == nil then
    status("Gui Needs To Be Open") --TODO remove menu entry when gui closes
    return
  end
  --song object
  local song = renoise.song()
  --get the last track (in case there are less than 12 sequencer tracks to copy devices from)
  local last_track =  song.sequencer_track_count
  if last_track > 12 then
    last_track = 12
  end
  --selected page in tool gui 
  local page = g_selected_page 
  --clear tag names that will be overwritten with track names
  
  
  --get fx pages[] node
  local xml_page = g_retrieved_xml.plugin_types[TYPE_EFFECT].pages 
  ---------------------------------------------
  --delete all current devices saved in the tag
  ---------------------------------------------
  for tag = last_track,1,-1 do
    for i = #xml_page[g_selected_page].tags[tag].plugs,1,-1 do
      --get name from xml object
      local name = tostring(xml_page[g_selected_page].tags[tag].plugs[i].plug_name)
      --delete using global function
      remove_plugin_node(TYPE_EFFECT,g_selected_page,tag,name)
    end
  end
 -- (was #g_retrieved_xml[page]["tag_names"], now last_track)
 -- for i = last_track,1,-1 do
 --   g_retrieved_xml[page]["tag_names"]:remove(i)
 -- end

  --loop tracks 1 to 12 in song to add devices
  --------------------------------------------
  for track_idx = 1,last_track do 
   -- print(track_idx)
    --call add all devices in track fn passing track index
    --we need to pass the tag_name that the plugs will be added to.
   -- local tag_name = tostring("tag"..track_idx)
    
    add_all_devices_in_track_to_tag(track_idx,track_idx)--,false)  --each track num corresponds to tag num
    
    --update button names from the track names
    local tag = track_idx
    local track_name = song:track(track_idx).name
    --truncate text to match button width
    track_name = string.sub(track_name,1,13)
    
    --add track name into xml document object as the tag name.  This way the user can name
    --the tracks in their XRNX (song file) to store/update tags at the same time 
    xml_page[page].tags[tag].name.value = track_name
   
   -- g_retrieved_xml[page]["tag_names"]:insert(track_idx, track_name)
    --update gui buttons with vb.views
    vb.views["tag "..tostring(track_idx)].text = track_name
  end
  
  --status
  status("Devices Added")
end

--------------------------------------------------------
--------------------------------------------------------
--------------------------------------------------------
--------------------------------------------------------




--4) 
--when calling multiple :remove([pos]) actions e.g. in a loop, call in reverse order
--as list indexes are update after each call.
--------------------------------------------------------- 
function remove_entry_from_tab(page_idx,tag_idx,item_idx,item_name)
---------------------------------------------------------
  
  --make sure all parameters are passed
  if (page_idx == nil) or (tag_idx == nil) or (item_idx == nil) then
    error("need valid parameters")
    return
  end
  
  --declare fx pages[] node object
  local xml_page = g_retrieved_xml.plugin_types[TYPE_EFFECT].pages
  
  --is item_idx too high?
  if item_idx > #xml_page[page_idx].tags[tag_idx].plugs then
    error("Choose A Valid Entry In Popup")
    return
  end

  --remove plug from xml
  remove_plugin_node(TYPE_EFFECT,page_idx,tag_idx,item_name) --(TYPE_EFFECT,PAGE_1,TAG_1,"Compressor")
  
  --(done in notifier) NOTE CHECK FOR BEST WAY--
  --update the main GUI 
  --update_load_popup()
  --remember to:
  --save_to_xml_file()
end 
  


--6)
--sorts all the entries in a specified tag (defaults to selected tag)
----------------------------------------------
function sort_tag_list(target_page,target_tag) 
----------------------------------------------

  --create a table to add the device properties into.
  local device_props = {}
  device_props.name = {}
  device_props.path = {}
 
  --use passed params or default values for page and tag
  local page = target_page or g_selected_page
  local tag = target_tag or g_selected_button
  
  --declare fx pages[] node
  local xml_page = g_retrieved_xml.plugin_types[TYPE_EFFECT].pages
 
  --loop all entries in xml and add to device_props table
  --------------------------
  for i = 1,#xml_page[page].tags[tag].plugs do
    device_props.path[i] = xml_page[page].tags[tag].plugs[i].plug_path.value
    device_props.name[i] = xml_page[page].tags[tag].plugs[i].plug_name.value
  end
  
  --local function to sort the device_props table (called directly below definition)
  -----------------------------------------------
  local function sort_plugs_by_display_names(tab)
  -----------------------------------------------
    
    --temporary table
    local temp = {}
    --an append_digit will be added to the end of each entry after an index number.  The append_digit tells us how many
    --digits to expect in the index (normal usage should be 1/2 as unlikely that 100+ plugs would be added to a tag)
    --this is done so that if the end of a plugin name is a number, it will not confused with the added index in later sorting.
    --Example for table entry 7 after loop:
    --"[7] =>  "TR5 TAPE ECHO71" --- '7' is the index, '1' is the append_digit telling us that 7 is a single digit number.
    local append_digit = 1
    --loop entries, adding original table position index
    for i = 1,#tab.name do
      --set the append digit to the number of digits of 'i'
      append_digit = #tostring(i)
      --capitalise the plug name for sorting (keeps later sort results 'alphabetical' rather than ASCII)
      temp[i] = tab.name[i]:upper()
      --add the loop index number 'i' and + 'append_digit' to specify number of digits 'i' has (1 digit for 5, 2 digits for 16 etc.)
      temp[i] = temp[i]..tostring(i)..tostring(append_digit)
    end
   
    -------------------------------------------------------------
    --sort the capitalised table alphabetically (Lua standard fn)
    -------------------------------------------------------------
    table.sort(temp)
    
    --now the temporary table of plug names is sorted alphabetically we retrieve the
    --index numbers that were appended to the ends of the string entries denoting their original table positions
    ------------------------------------------------------------------------------------------------------------
    local index_table = {}
    for i = 1,#temp do
      --get append_digit
      local entry = temp[i]
      --get the append_digit which is the last character of the string.
      local append_digit = entry:sub(#entry)
      --get the index (appended number minus the append digit).  Add it to the index_table
      index_table[i] = entry:sub(#entry - append_digit, #entry-1)
    end
    
    --make a new 'sorted table' to hold the alphabetically sorted entries
    local sorted_tab = {}
    sorted_tab.name = {}
    sorted_tab.path = {}
    --loop to re-index original table
    for i = 1,#temp do
      --by indexing index_table we get the plug names back in alphabetical order.
      local index = tonumber(index_table[i])
      --now build the sorted table using the original saved indexes
      table.insert(sorted_tab.name,tab.name[index])
      table.insert(sorted_tab.path,tab.path[index])--make sure to add paths aswell
    end
    --return the sorted table
    return sorted_tab
  end 
  -------------------------- end local fn
  --------------------------
  
  --re-order device_props{} alphabetically, using above fn
  device_props = sort_plugs_by_display_names(device_props)

  --clear xml entries from global object
  for i = 1,#xml_page[page].tags[tag].plugs do
    local name = device_props.name[i]
    remove_plugin_node(TYPE_EFFECT,page,tag,name) --removes path and name
    
   -- device_props.name = xml_page[page].tags[tag].plugs[i].plug_path.value
   -- device_props.name = xml_page[page].tags[tag].plugs[i].plug_name.value
  end
  
  print(#device_props.name)
  
  --re-add entries name/path from  now alphabetical device_props table
  for i = 1,#device_props.name do
    insert_new_plugin_node(TYPE_EFFECT,page,tag,device_props.name[i],device_props.path[i])-- (TYPE_EFFECT,PAGE_1,TAG_1,"Compressor","Audio/Effects/Native/Compressor")
  end
  
  --re-save xml
  save_to_xml_file()
  --update the popup to show the updated list
  update_load_popup()  
end
---





